/*
	KOHONEN SELF-ORGANIZING NEURAL NETWORK
*/

#include <graphics.h>
#include <conio.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <ctype.h>
#include <string.h>
#include <time.h>
#include <dos.h>

#define TOP    40
#define BOTTOM 440
#define LEFT   120
#define RIGHT  520
#define PI     3.141592654
#define TRUE 1
#define FALSE 0


typedef float *PFLOAT;     /* pointer to a real    */
typedef PFLOAT VECTOR;     /* a vector: one colums */
typedef PFLOAT *MATRIX;    /* a matrix: two colums */

/* Prototipos de las funciones utilizadas */

void AllocateVector(VECTOR *vector, int nCols);
void AllocateColumns(PFLOAT Matrix[], int nRows, int nCols);
void AllocateMatrix(MATRIX *pmatrix, int nRows, int nCols);
void FreeMatrix(MATRIX Matrix, int nRows);
void help(void);
void initweights(int x, int y, int R);
void readpatterns(char *szPattern);
void writeweights(char *szWeights);
void writeactivations(char *szResults);
int  findwinner(int p);
void normalizeweights(int j);
void ResetScreen (void);
void PutPoint (float x, float y, int color, int size);

/* Definicion de las matrices de almacenamiento dinamico */

MATRIX   out0;            /* input layer */
MATRIX   out2;            /* output layer  */
MATRIX   Weights;         /* weights for imput to output layer */
MATRIX   Weights_Muertos;  /* weights for imput to output layer no modificados */
unsigned nInputNodes;
unsigned nOutputNodes;
unsigned nPatterns;
long int nIterations;
unsigned p;               /* indice del numero de patron */
float    alfa = 1;        /* parametro de aprendizaje actual */
unsigned q;               /* cuenta de intentos */
long int t;               /* periodo de tiempo en segmento actual  */
float    u,v;

/* ------------------------------------------------------------------------ */
void main(int argc, char *argv[])
/* ------------------------------------------------------------------------ */
{
float        alfa1;             /* alfa(t) = alfa1*(1 - t/alfa2) */
float        alfa2;
register int i;                 /* index hidden layer   */
register int j;                 /* index input layer   */
int          r;                 /* index output layer */
int          n;                 /* numero de capas intermedias */
int          nRuns;             /* number of input nodes   */
int          T2;                /* show network state every T2 iterations */
FILE         *fpRun;            /* run file  */
char         szResults[66];
char         szPattern[66];
char         szWeightsOut[66];
int          Winner;
float        sum;
/* Variables utilizadas para pintar los puntos */
int  Ganadora_p, Ganadora_m, c_aux, fin = FALSE, c_leido;
int c_Ganadora_p, c_Muerta, c_Resto;
char aux_s1[50], aux_s2[10];
float error_c = 0.001;

	/* Read arguments from DOS command line.*/

	if (argc != 3)
		help();

	T2 = atoi(argv[1]);

	/* open run file for reading, */

	if ((fpRun = fopen(argv[2], "r")) == NULL)
	{
		fprintf(stderr,"Can't open file %s \n", argv[2]);
		exit(1);
	}

	fscanf(fpRun, "%d",&nRuns); /* scan for numbers of runs, */

	/*******************************************/
	/* Comienzo del bucle de trabajo Principal */
	/*******************************************/

	for (r=0; r<nRuns; r++)
	{
	/* Read and parse the run specification line to obtain */
	/* information about this network. */

	/******************************************************/
	/* Los parametros del fichero ejecucion son :		   */
	/*                                                    */
	/*	Fichero de resultados finales			               */
	/*	Fichero de patrones de entrenamiento               */
	/*	Fichero de pesos de salida    		               */
	/*	Numero de patrones 				                     */
	/* Numero de iteraciones                              */
	/*	Numero de nodos de entrada			                  */
	/* Numero de nodos de salida   			               */
	/*	Parametros de aprendizaje: alfa1 y alfa2           */
	/******************************************************/

		fscanf(fpRun, "%s %s %s %d %u %d %d %f %f",
			szResults,                /* output results file */
			szPattern,                /* pattern input file  */
			szWeightsOut,             /* final weights output filt  */
			&nPatterns,               /* number of patterns to lea  */
			&nIterations,             /* number of iterations the data */
			&nInputNodes,             /* number of input nodes   */
			&nOutputNodes,            /* number Of Output nodeS  */
			&alfa1,
			&alfa2);                  /* learning rates          */

		/* Allocate dynamic storage for nodes and patterns. */

		AllocateMatrix(&out0, nPatterns, nInputNodes);
		AllocateMatrix(&out2, nPatterns, nOutputNodes);
		AllocateMatrix(&Weights, nOutputNodes, nInputNodes);

        AllocateMatrix(&Weights_Muertos, nOutputNodes, 1);
        for (j=0; j<nOutputNodes; j++)
		    	{Weights_Muertos[j][0] = TRUE;
				} /* de j */

		/*  Create the initial Weights matrices   */
		initweights(20, 20, 5);

		/* Read in all patterns to be learned, */
		readpatterns(szPattern);

		/* Inicializar el tiempo */
		t = 0;

		/********************************************************************/
		/* Partes basicas de las operaciones en el lapso de tiempo t;       */
		/*	este bucle interior consta de T1 lapsos de tiempo;               */
		/*	Por cada lapso de tiempo:                                        */
		/*		.- un patron se presenta.                                     */
		/*		.- un ganador se selecciona.                                  */
		/*		.- los pesos de los vecinos son adaptados.                    */
		/*                                                                  */
		/* En cada iteracin q se presenta el conjunto completo de patrones */
		/********************************************************************/

		/* Se muestra el estado inicial de la red */

		ResetScreen ();
		setcolor (BLUE);
		for (i=0; i<nPatterns; i++)
			PutPoint (out0[i][0], out0[i][1], LIGHTRED, 4);
		for (i=0; i<nOutputNodes; i++)
		{
			PutPoint (Weights[i][0], Weights[i][1], LIGHTBLUE, 2);
			if (i != 0)
				line (LEFT + u*400, BOTTOM - v*400,
						LEFT + Weights[i][0]*400, BOTTOM - Weights[i][1]*400);
			u = Weights[i][0];
			v = Weights[i][1];
		}
		line (LEFT + u*400, BOTTOM - v*400,
				LEFT + Weights[0][0]*400, BOTTOM - Weights[0][1]*400);
		delay (1000);

		for (q=0; q<nIterations && alfa>0; q++)   /* numero de iteraciones */
		{
			for (p = 0; p < nPatterns; p++)        /* numero de patrones   */
			{
			/* calcula la activacion de la capa de salida  */

				for (j=0; j<nOutputNodes; j++)
				{
					sum = 0;
					for (i=0; i<nInputNodes; i++)
						sum += Weights[j][i] * out0[p][i];

					/* Compute output. */
					out2[p][j] = sum;
				} /* de j */

				/* buscamos el nodo ganador  */

				Winner = findwinner(p);

                Ganadora_m = Winner;

				/* Ajustamos los pesos de las vecinas de la ganadora */
				/* La vecindad es lineal y de radio 1 */

				/* el coeficiente de apredizaje es una funcion lineal discreta
					de lapso de tiempo t */

				alfa = alfa1 * ( 1 - ((float)t/(float)alfa2) );

				/* Modificar los pesos de los vecinos proximos al ganador */

				if ((j = Winner - 1) < 0)
					j = nOutputNodes - 1;
				for (i=0; i<nInputNodes; i++)
					{Weights[j][i] += alfa*(out0[p][i]-Weights[j][i]);
                    Weights_Muertos[j][0] = FALSE;
                    }

				if ((j = Winner + 1) == nOutputNodes)
					j = 0;
				for (i=0; i<nInputNodes; i++)
					{Weights[j][i] += alfa*(out0[p][i]-Weights[j][i]);
                    Weights_Muertos[j][0] = FALSE;
                    }


				for (i=0; i<nInputNodes; i++)
					{Weights[Winner][i] += alfa*(out0[p][i]-Weights[Winner][i]);
                    Weights_Muertos[Winner][0] = FALSE;
                    }


			/*} /* for de p*/
            /*
			t++;*/
			if ((fin == FALSE) || ((t % T2 == 0) && (p == nPatterns - 1)))
			{   c_Ganadora_p = 0;
                c_Muerta = 0;
                c_Resto = 0;

				setfillstyle (SOLID_FILL, BLACK);
				bar (LEFT-5, TOP-40, RIGHT+5, BOTTOM+30);
				for (i=0; i<nPatterns; i++)
					PutPoint (out0[i][0], out0[i][1], LIGHTRED, 4);
				for (i=0; i<nOutputNodes; i++)
				{   /*-----------------------CODIGO DE DAVID -------------*/
                    Ganadora_p = FALSE;

                	for (c_aux=0; c_aux < nPatterns; c_aux++)
				    	{
                        if (((Weights[i][0] >= out0[c_aux][0] - error_c) &&
                           (Weights[i][0] <= out0[c_aux][0] + error_c))
                            &&
                            ((Weights[i][1] >= out0[c_aux][1] - error_c) &&
                            (Weights[i][1] <= out0[c_aux][1] + error_c)))
                               {Ganadora_p = TRUE;
                               }
                        }
                    if (Weights_Muertos[i][0] == TRUE)
                       {PutPoint (Weights[i][0], Weights[i][1], YELLOW, 2);
                       c_Muerta++;
                       }
                    else
                        if (Ganadora_p == TRUE)
                            {PutPoint (Weights[i][0], Weights[i][1], GREEN, 2);
                            c_Ganadora_p++;
                            }
                        else
                            {PutPoint (Weights[i][0], Weights[i][1], LIGHTBLUE, 2);
                            c_Resto++;
                            }

                    /*----------------------------------------------------*/
                    /*PutPoint (Weights[i][0], Weights[i][1], LIGHTBLUE, 2);*/
					if (i != 0)
						line (LEFT + u*400, BOTTOM - v*400,
								LEFT + Weights[i][0]*400, BOTTOM - Weights[i][1]*400);
					u = Weights[i][0];
					v = Weights[i][1];
				}
				line (LEFT + u*400, BOTTOM - v*400,
						LEFT + Weights[0][0]*400, BOTTOM - Weights[0][1]*400);
/*				delay (500);*/
                setcolor (LIGHTGREEN);
                PutPoint (Weights[Ganadora_m][0], Weights[Ganadora_m][1], WHITE, 3);
                strcpy(aux_s1,"Neuronas ganadoras: ");
                itoa(c_Ganadora_p,aux_s2,10);
                strcat(aux_s1, aux_s2);
                outtextxy (210, 0, aux_s1);
                strcpy(aux_s1,"Neuronas muertas: ");
                itoa(c_Muerta,aux_s2,10);
                strcat(aux_s1, aux_s2);
                outtextxy (210, 10, aux_s1);
                strcpy(aux_s1,"Resto de neuronas: ");
                itoa(c_Resto,aux_s2,10);
                strcat(aux_s1, aux_s2);
                outtextxy (210, 20, aux_s1);

                if (fin == FALSE)
                   {outtextxy (210, 450, "Pulse una tecla para la traza.");
                   outtextxy (210, 460, "Pulse F para continuar.");
                   c_leido = getch();
                   if ((70 == c_leido) || (c_leido == 102))
                      {fin = TRUE;
                      }
                   }
               else
				   {delay (500);
                   }
               setcolor (BLUE);
			}
            } /* for de p */
            t++;
		} /* for de q */

		/* Escribe los pesos finales */

		writeweights(szWeightsOut);

		/* Escribe los valores de activacion finales */

		writeactivations(szResults);

		/* Libera la memoria. */

		FreeMatrix(out0, nPatterns);
		FreeMatrix(out2, nPatterns);
		FreeMatrix(Weights, nOutputNodes);
        FreeMatrix(Weights_Muertos, nOutputNodes);

	} /* for de r (numero de ejecuciones) */

	fclose(fpRun);   /* Cerramos el fichero de ejecuciones */
	setcolor (LIGHTGREEN);
	outtextxy (210, 470, "Entrenamiento finalizado");
	getch ();
	closegraph ();
} /* de main */

/* ------------------------------------------------------------------------ */
void help()
/* ------------------------------------------------------------------------ */
{
	fprintf(stderr, "Usage: tsp n runfilename\n");
	fprintf(stderr, " n => report network state every n iterations\n");
	exit(1);
}

/* ------------------------------------------------------------------------ */
int findwinner(int p)
/* ------------------------------------------------------------------------ */
{
int Winner;
int i;
int j;
float EuclideanDistance, d, MinimunDistance = 0;

	for (j = 0 ; j < nOutputNodes; j++)
	{
		EuclideanDistance = 0;
		for (i = 0 ; i < nInputNodes; i++)
		{
			d = out0[p][i] - Weights[j][i];
			EuclideanDistance += d * d;
		}
		if (j==0 || EuclideanDistance <= MinimunDistance)
		{
			MinimunDistance = EuclideanDistance;
			Winner = j;
		}
	}
	return Winner;
} /* de findwinner */

/* ------------------------------------------------------------------------ */
void initweights(int x, int y, int R)
/* ------------------------------------------------------------------------ */
/* 	Se inicializan los pesos en una circunferencia de centro (x,y) y de   */
/* radio R.                                                                 */
/* ------------------------------------------------------------------------ */
{
int j;
float da, angle;

	da = 2*PI/nOutputNodes;
	angle = 0;
	for (j=0; j<nOutputNodes; j++)
	{
		Weights[j][0] = R*cos(angle) + x;
		Weights[j][1] = R*sin(angle) + y;
		normalizeweights(j);
		angle += da;
	}
}

/* ------------------------------------------------------------------------ */
void normalizeweights(int j)
/* ------------------------------------------------------------------------ */
{
int i;

	for (i = 0 ; i < nInputNodes; i++)
		Weights[j][i] /= 40;
}

/* ------------------------------------------------------------------------ */
void readpatterns(char *szPattern)
/* ------------------------------------------------------------------------ */
{
int i;
FILE *fpPattern;    /*  fichero de los patrones de entrada  */

	if (( fpPattern = fopen(szPattern,"r")) == NULL)
	{
		fprintf(stderr, "can't open file %s\n",szPattern);
		exit(1);
	}
	/*  Leer los pesos de las conexiones input : output */

	for (p = 0 ; p < nPatterns; p++)
	{
		for (i = 0 ; i < nInputNodes; i++)
		{
			float r;
			if (fscanf(fpPattern, "%f", &r) != 1)
				goto ALLPATTERNSREAD;
			/* Normaliza los Patrones de Entrada */
			out0[p][i] = r/40.0;
		}
	}

	ALLPATTERNSREAD:
	fclose(fpPattern);

	if (p < nPatterns)
	{
		fprintf(stderr,"%d  de  %d patrones leidos \n",p,nPatterns);
		nPatterns = p;
	}
}

/* ------------------------------------------------------------------------ */
void writeweights(char *szWeightsOut)
/* ------------------------------------------------------------------------ */
{
int i;
int j;
FILE * fpWeightsOut;

	if (( fpWeightsOut = fopen(szWeightsOut,"w")) == NULL)
	{
		fprintf(stderr, "can't open file %s\n",szWeightsOut);
		exit(1);
	}
	for (j = 0 ; j < nOutputNodes; j++)
		for (i = 0 ; i < nInputNodes; i++)
			fprintf(fpWeightsOut, "%f\n",Weights[j][i]);
	fclose(fpWeightsOut);
}

/* ------------------------------------------------------------------------ */
void writeactivations(char *szResults)
/* ------------------------------------------------------------------------ */
{
int j;
FILE * fpResults;

	if (( fpResults = fopen(szResults,"w")) == NULL)
	{
		fprintf(stderr, "can't open file %s\n",szResults);
		fpResults = stderr;
	}
	for (p = 0 ; p < nPatterns; p++)
	{
		for (j = 0 ; j < nOutputNodes; j++)
			fprintf(fpResults, "%f\n",out2[p][j]);
	}
	fclose(fpResults);
}

/* ------------------------------------------------------------------------ */
void AllocateVector(VECTOR *Vector, int nCols)
/* ------------------------------------------------------------------------ */
{
	if ((*Vector = (VECTOR) calloc(nCols, sizeof(float))) == NULL)
	{
		fprintf(stderr, " Sorry Not enough memory for nodes !\n");
		exit(1);
	}
}

/* ------------------------------------------------------------------------ */
void AllocateColumns(PFLOAT Matrix[], int nRows, int nCols)
/* ------------------------------------------------------------------------ */
{
int i;

  for (i = 0; i < nRows; i++)
	 AllocateVector(&Matrix[i], nCols);
}

/* ------------------------------------------------------------------------ */
void AllocateMatrix(MATRIX *Pmatrix,int nRows, int nCols)
/* ------------------------------------------------------------------------ */
{
	if ((*Pmatrix = (MATRIX) calloc(nRows, sizeof(PFLOAT))) == NULL)
	{
		fprintf(stderr, " Sorry Not enough memory for nodes !\n");
		exit(1);
	}
	AllocateColumns(*Pmatrix, nRows, nCols);
}

/* ------------------------------------------------------------------------ */
void FreeMatrix(MATRIX Matrix, int nRows)
/* ------------------------------------------------------------------------ */
{
int i;

	for (i = 0; i < nRows; i++)
	free(Matrix[i]);
	free(Matrix);
}

/* ------------------------------------------------------------------------ */
void ResetScreen (void)
/* ------------------------------------------------------------------------ */
{
int gdriver = DETECT, gmode, errorcode;

	initgraph(&gdriver, &gmode, "");
	errorcode = graphresult();
	if (errorcode != grOk)
	{
		printf("Graphics error: %s\n", grapherrormsg(errorcode));
		printf("Press any key to halt:");
		getch();
		exit(1);
	}
}

/* ------------------------------------------------------------------------ */
void PutPoint (float x, float y, int color, int size)
/* ------------------------------------------------------------------------ */
{
int dx, dy;
	if ((x >= 0) && (x <= 1) && (y >= 0) && (y <= 1))
		for (dx=-size; dx<=size; dx++)
			for (dy=-size; dy<=size; dy++)
				putpixel (LEFT + x*400 + dx, BOTTOM - y*400 + dy, color);
}
